/* ExpPyRE.c -- support for perl-compatible regular expressions */
/* Copyright (C) 1998 Michael P. Reilly, All rights reserved
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: ExpPyRE.c,v $
 * Revision 1.15  2000/11/04 01:35:57  arcege
 * Merge from EXPECTPY_1_8_3; SF-bug#121064
 *
 * Revision 1.14.2.1  2000/11/03 12:37:43  arcege
 * Bug#121064: Import "pre" instead of "re"(sre) in Python 2.0
 *
 * Revision 1.14  1999/07/05 11:57:37  arcege
 * Make code more compilable (K&R)
 *
 * Revision 1.13  1999/04/21 16:57:22  arcege
 * If Python is not compiled with threads; then fatal error
 *
 * Revision 1.12  1999/04/20 18:14:30  arcege
 * Use new common malloc interface;  store thread state with regular expression
 * structure to make ExpectPy Python-thread safe;  fix CVS log string.
 *
 *
 */

#include "ExpPycnf.h"
#include <stdio.h>
#include <Python.h>
#include <abstract.h>
/* 
#define EXP_AVOID_INCLUDING_TCL_H
#include <expect.h>
 */

#include "ExpPyMem.h"
#include "exppyre.h"
#include "ExpPyRE.h"

PyObject *pcre_compile_obj;

struct regexp *
TclRegComp(exp)
  char *exp;
  { return NULL; }

/* free up the regexp, including the enclosed re */
void
ExpPyRE_Free(prog)
  struct regexp *prog;
  {
    /*fprintf(stderr, "Freeing regular expression\n");*/
    Py_XDECREF(prog->real_re->re);
    prog->real_re->re = NULL;
    /* what should be done with real_re->tstate ? */
    ExpPy_free(prog->real_re);
    prog->real_re = NULL;
    ExpPy_free(prog);
  }

PyObject *
ExpPyRE_re(prog)
  struct regexp *prog;
  {
    return prog->real_re->re;
  }

/* create a new regexp "object" */
struct regexp *
ExpPyRE_New(exp)
  PyObject *exp;
  { struct regexp *new;
    int size = sizeof(struct regexp);

    /*fprintf(stderr, "Creating new re with %p\n", exp);*/
    if ((new = ExpPy_malloc(struct regexp)) == NULL)
      return NULL;

    /*fprintf(stderr, "TclRegCreate(%p) => %p\n", exp, new);*/
    if ((new->real_re = ExpPy_malloc(struct ExPyRE)) == NULL) {
      ExpPy_free(new);
      return NULL;
    }
#ifdef WITH_THREAD
    new->real_re->tstate = PyThreadState_Get();
#else
    new->real_re->tstate = NULL;
#endif
    Py_INCREF(exp);
    new->real_re->re = exp;
    return new;
  }

/* create an regexp object based on a pattern, compile the pattern with
   PCRE.
 */
struct regexp *
ExpPyRE_Create(exp)
  PyObject *exp;
  { struct regexp *new;
    PyObject *pcre_obj, *args;

    if ((args = Py_BuildValue("(Oi)", exp, 0)) == NULL)
      return NULL;
    if ((pcre_obj = PyEval_CallObject(pcre_compile_obj, args)) == NULL) {
      Py_DECREF(args);
      return NULL;
    }
    Py_DECREF(args);
    if ((new = ExpPyRE_New(pcre_obj)) == NULL) {
      Py_DECREF(pcre_obj);
      return NULL;
    }

    return new;
  }

/* this is a replacement of the Tcl low-level regular expression
   execution.  The regexp contains a re.compile object which is matched
   with the passed string.  The return value of the match (which is
   a "match" object is assigned to the attribute "lastmatch" in the
   re object.
 */
int
TclRegExec(prog, string, start)
  struct regexp *prog; char *string; char *start;
  { PyObject *match, *match_meth, *args;
    PyObject *regs, *t;
    PyObject *real_re;
    int result, i, l, s, e;

    real_re = prog->real_re->re;
#ifdef WITH_THREAD
    PyEval_RestoreThread(prog->real_re->tstate);
#endif
    if ((match_meth = PyObject_GetAttrString(real_re, "match")) == NULL)
      return 0;
    args = Py_BuildValue("(s)", start);

    /*fprintf(stderr, "Matching %p against \"%s\"\n", prog, start);*/
    match = PyEval_CallObject(match_meth, args);
    /*PyObject_Print(match, stderr, 0); fprintf(stderr, "\n");*/
    if (match == NULL || match == Py_None)
      result = 0;
    else {
      result = 1;
      /* assign the startp and endp arrays from the groups regs */
      regs = PyObject_GetAttrString(match, "regs");
      l = PyList_Size(regs);
      for (i = 0; i < l; i++) {
        t = PyList_GetItem(regs, i);
        PyArg_Parse(t, "(ii)", &s, &e);
        prog->startp[i] = &start[s];
        prog->endp[i] = &start[e];
      }
      Py_DECREF(regs);
    }

    if (PyObject_HasAttrString(real_re, "lastmatch"))
      PyObject_DelAttrString(real_re, "lastmatch");
    if (match != NULL)
      PyObject_SetAttrString(real_re, "lastmatch", match);

    /*Py_DECREF(match);*/
    Py_DECREF(args);
    Py_DECREF(match_meth);
#ifdef WITH_THREAD
    PyEval_SaveThread();  /* this should be prog->real_re->tstate */
#endif
    return result;
  }

/* this is still just a stub for now */
void
TclRegSub(prog, buffer, dest)
  struct regexp *prog;
  char *buffer, *dest;
  {
    strcpy(dest, buffer);
  }

/* These are just stubs for the Expect library */
char *Tcl_ErrnoMsg (err)
  int err;
  { return ""; }
char *TclGetRegError ()
  { return NULL; }
void TclRegError(message)
  char *message;
  { return; }

/* import the re (or pre for Py2.0) module, we only need the compile method */
void
initExpPyRE()
  { PyObject *mod;

#if PYTHON_API_VERSION >= 1009
    if (mod = PyImport_ImportModule("pre"))
#else
    if (mod = PyImport_ImportModule("re"))
#endif
      pcre_compile_obj = PyObject_GetAttrString(mod, "compile");
    Py_XDECREF(mod);
  }

