/*  expect.c  -- Imbed the expect library into a python extension */
/* Copyright (C) 1998 Michael P. Reilly, All rights reserved
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 * $Log: expect.c,v $
 * Revision 1.55  2000/11/04 01:35:57  arcege
 * Merge from EXPECTPY_1_8_3; SF-bug#121064
 *
 * Revision 1.54  2000/05/14 17:28:41  arcege
 * Return to exp_expectv function if interrupted
 *
 * Revision 1.53  2000/02/22 13:15:33  arcege
 * Move stty object to stty.c
 *
 * Revision 1.52  2000/02/17 13:25:59  arcege
 * Get stty functionality from stty.c
 *
 * Revision 1.51  1999/11/25 02:21:28  arcege
 * merge from 1.8.3
 *
 * Revision 1.50  1999/10/28 10:58:07  arcege
 * Allow file descriptor to be given to spawnfile
 *
 * Revision 1.49  1999/09/02 11:30:27  arcege
 * Move the stty object definition earlier in the file for clearer code
 *
 * Revision 1.48  1999/09/02 11:25:28  arcege
 * The major change is to take the stty() method and convert it to a callable
 * object with assignable attributes.
 * Fix a potential bug where the given command to new_ExpectObject() could
 * overwrite memory.
 * Add code for new, more efficient API functions.
 *
 * Revision 1.47  1999/08/29 13:16:53  arcege
 * Fix some typos and missing casts while porting to AIX 4.x.  Add
 * conditionals for termios values.
 *
 * Revision 1.46  1999/07/19 12:31:03  arcege
 * Add some initial form of interrupt sequence to interact()
 *
 * Revision 1.45.4.1  2000/11/02 13:05:07  arcege
 * Bug#121064: Import "pre" instead of "re"(sre) in Python 2.0
 *
 * Revision 1.45.2.2  1999/11/24 14:10:52  arcege
 * Some do not have cfsetspeed()
 *
 * Revision 1.45.2.1  1999/11/24 13:15:50  arcege
 * FreeBSD does not have buad rates accessible from the termios structure.
 * FreeBSD causes the process to terminate prematurely; change the expect
 * method to return EOFError if that happens.
 *
 * Revision 1.45  1999/07/05 15:22:42  arcege
 * Add helper macros
 *
 * Revision 1.44  1999/07/05 12:07:23  arcege
 * Bug where settings object could receive a NULL on setattr
 *
 * Revision 1.43  1999/07/05 11:50:02  arcege
 * Make code more compilable (K&R)
 *
 * Revision 1.42  1999/07/04 23:34:50  arcege
 * Do not change all casts
 *
 * Revision 1.41  1999/07/04 22:05:08  arcege
 * Add code for Tcl 8.1 later
 *
 * Revision 1.40  1999/06/13 14:02:20  arcege
 * Linux handles extended CBAUD differently
 *
 * Revision 1.39  1999/04/25 15:33:38  arcege
 * Add doc string for interact.  Remove trailing newlines from other doc
 * strings.
 *
 * Revision 1.38  1999/04/25 13:21:31  arcege
 * More doc strings
 *
 * Revision 1.37  1999/04/25 13:17:59  arcege
 * Code for stty() method returning a tuple or a dict
 *
 * Revision 1.36  1999/04/24 20:14:27  arcege
 * Add keyword arguments and baud setting to stty() method
 *
 * Revision 1.35  1999/04/23 21:00:41  arcege
 * Add stty() method
 *
 * Revision 1.34  1999/04/23 12:40:18  arcege
 * Add controls for spawn tty initialization
 *
 * Revision 1.33  1999/04/23 12:08:47  arcege
 * Raise TimeoutError inside interact function, as appropriate
 *
 * Revision 1.32  1999/04/20 19:30:11  arcege
 * Fix string type Expect variable memory management
 *
 * Revision 1.31  1999/04/20 18:12:10  arcege
 * Use new common malloc interface;  make the module functions and object
 * methods Python-thread safe.  Normalize some exception value strings.  Fix
 * bug where when settings.reset is given an invalid attribute name a core
 * dump might occur.
 *
 * Revision 1.30  1999/04/16 11:08:26  arcege
 * Abstract the coding to access the pyexp_variable map structure, this
 * allows for easier changes to the searching algorithm later.  Add the
 * concept of "factory default settings" to the settings object, with
 * a method for an interface.  Add a method to return the Expect variable
 * values as a dictionary.
 * Add more __doc__ strings.
 *
 * Revision 1.29  1999/04/13 10:58:48  arcege
 * Handle Python 1.4 API; explicitly cast assignments (which should not
 * need them)
 *
 * Revision 1.28  1999/04/11 18:11:01  arcege
 * Remove redundant/old code
 *
 * Revision 1.27  1999/04/11 12:31:21  arcege
 * Add rudimentary interact method/function
 *
 * Revision 1.26  1999/04/10 16:55:29  arcege
 * Fix castings
 *
 * Revision 1.25  1999/04/10 16:48:14  arcege
 * Move definitions and prototypes into header files.  Move version number
 * into patchlevel.h.
 *
 * Revision 1.24  1998/12/31 15:35:55  arcege
 * Fix bug where proper integers where not allowed in set method
 *
 * Revision 1.23  1998/12/31 12:47:48  arcege
 * Fix bug where accessing the "buffer" member of the spawn object causes a
 * segmentation fault.
 *
 * Revision 1.22  1998/12/31 12:40:32  arcege
 * Change tabs to spaces
 *
 * Revision 1.21  1998/09/11 10:34:34  arcege
 * Add Python 1.4 memory allocation
 *
 * Revision 1.20  1998/08/24 01:29:02  arcege
 * Use PYTHON_API_VERSION instead of PYTH_VERS; remove PYTH_VERS
 *
 * Revision 1.19  1998/08/19 11:42:53  arcege
 * Fix some MAJOR reference count bugs from recent coding
 *
 * Revision 1.18  1998/08/16 17:29:47  arcege
 * Pass regexp object in callbacks for exp_compiled clauses
 *
 * Revision 1.17  1998/08/12 22:55:09  arcege
 * Add is_alive predicate
 *
 * Revision 1.16  1998/08/11 11:55:53  arcege
 * Add exp_continue support; fix build warning in pyex_set function
 *
 * Revision 1.15  1998/08/10 12:38:19  arcege
 * Allow for hashable return values in expect method
 *
 * Revision 1.14  1998/08/05 12:34:59  arcege
 * Add Python 1.4 support; includes online functions in 1.5
 *
 * Revision 1.13  1998/08/01 12:09:00  arcege
 * Fix prototype
 *
 * Revision 1.12  1998/08/01 11:41:39  arcege
 * Add TclRE support
 *
 * Revision 1.11  1998/07/30 11:18:36  arcege
 * Add comments
 *
 * Revision 1.10  1998/07/29 12:34:51  arcege
 * Add "re_type" value for module at compile-time
 *
 * Revision 1.9  1998/07/28 00:14:26  arcege
 * Rename "config.h" to "ExpPycnf.h"
 *
 * Revision 1.8  1998/07/27 12:21:02  arcege
 * Copyright information
 *
 * Revision 1.7  1998/07/27 11:27:03  arcege
 * Deactivate loguser
 *
 * Revision 1.6  1998/07/26 17:58:51  arcege
 * Add support for creating a spawn object from an open file, not a command.
 * When closing one of these kinds of spawn objects, do not close the file
 * descriptor; assume it is owned by something else.  Rework the repr()
 * method in the same way.  Initialize the module with open spawns for devtty,
 * stdin/stdout and stderr (from the values in the sys module).
 * When a FULLBUFFER error has been detected, store the buffer as with EOF
 * errors.
 * Made some of the functions static.
 *
 */

#include "ExpPycnf.h"
#include "patchlevel.h"
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <errno.h>
#include <fcntl.h>
#if HAVE_UNISTD_H
#  include <unistd.h>
#endif
#if HAVE_SYS_WAIT_H
#  include <sys/wait.h>
#endif
#include <termios.h>

#include <Python.h>
#ifndef  HAVE_TCL
#define EXP_AVOID_INCLUDING_TCL_H
#endif
#include <expect.h>
#include <abstract.h>

#include "ExpPyMem.h"

/* external functions for regular expression support */
#ifdef HAVE_TCL
#include "ExpPyTcl.h"
#else
#ifdef HAVE_PCRE
#include "ExpPyRE.h"
#endif
#endif /* HAVE_TCL */

extern PyObject *pyexp_interact(PyObject *, PyObject *);
extern char pyexp_interact__doc__[];

/* exceptions */
static PyObject *Expect_Error;
static PyObject *Expect_FullBufferError;
/* used in pyexp_interact in ExpPyint.c */
PyObject *Expect_TimeoutError;

/* module variables */
static PyObject *version, *settings;
static char copyright[] =
  "Copyright (C) 1998 Michael P. Reilly, All rights reserved";
static PyObject *tty_spawn_id;

/* constants for set() command data types */
#define PYEXP_VAR_BOOL 001
#define PYEXP_VAR_INT  002
#define PYEXP_VAR_STR  003

/* code in stty.c */
extern PyObject *Stty_new();

/* spawn object structure */
typedef struct {
  PyObject_HEAD
  int id;      /* currently the file descriptor */
  int pid;     /* 0, if unspawned object */
  int closed;
  int Expcontinue;      /* 1, if a clause signals to continue */
  PyObject *command;
  PyObject *pty;     /* current device file (or symbolic name for unspawned) */
  PyObject *buffer;  /* current buffer; set on EOFError or
                        FullBufferError exceptions */
  PyObject *match;
  PyObject *sttyobj;
} ExpectObject;

staticforward PyTypeObject ExpObj_Type;
#define is_ExpObj(v)        ((v)->ob_type == &ExpObj_Type)
#define ExpObj_id(v)        (((ExpectObject *)(v))->id)
#define ExpObj_pid(v)       (((ExpectObject *)(v))->pid)
#define ExpObj_closed(v)    (((ExpectObject *)(v))->closed)
#define ExpObj_continue(v)  (((ExpectObject *)(v))->Expcontinue)
#define ExpObj_command(v)   (((ExpectObject *)(v))->command)
#define ExpObj_pty(v)       (((ExpectObject *)(v))->pty)
#define ExpObj_buffer(v)    (((ExpectObject *)(v))->buffer)
#define ExpObj_match(v)     (((ExpectObject *)(v))->match)
#define ExpObj_sttyobj(v)   (((ExpectObject *)(v))->sttyobj)

/* This will only return false when wait has been called to clean up this
   pid (after the process has exited).
 */
static int
pyexp_alive(pid)
  pid_t pid;
  { int rc;
    Py_BEGIN_ALLOW_THREADS
    rc = kill(pid, 0);
    Py_END_ALLOW_THREADS
    return (rc > -1);
  }

/* close the file descriptor and wait for the specified process */
static long
pyexp_close(fd, pid)
  int fd;
  pid_t pid;
  { long status;
    pid_t rc;

    Py_BEGIN_ALLOW_THREADS
    close(fd);
    rc = waitpid(pid, (int *)&status, 0);
    Py_END_ALLOW_THREADS
    if (rc <= 0)
      return (long)-1;
    return status;
  }

/* Handle exp_case structures */
static struct exp_case *
new_exp_cases(num)
  int num;
  {
    return ExpPy_calloc(struct exp_case, num);
  }

/* Free the exp_case array, including dereferencing exp_compiled re's */
static void
free_exp_cases(cases, num)
  struct exp_case *cases;
  int num;
  { int i;

#ifdef HAVE_PCRE
    for (i = 0; i < num; i++) {
      switch (cases[i].type) {
        case exp_compiled:
        case exp_regexp:
          if (cases[i].re != NULL)
            ExpPyRE_Free(cases[i].re);
          break;
        default: break;
      }
    }
#endif /* HAVE_PCRE */
    ExpPy_free((void *)cases);
  }

/*
  Analyze a tuple and assign the given exp_case struct appropriately.
  This assumes that the exp_case struct has been initialized.
  Return value:
    -1   - raise an exception upon returning, Python error already set
    0    - exit calling loop prematurely (found EXP_END early)
    ANY  - assignment succeeded
 */
static int
assign_exp_case(ecase, slice, callbacks)
  struct exp_case *ecase;
  PyObject *slice, *callbacks;
  { PyObject *obj, *cb_key;
    PyObject *cb_item;
    PyObject *pattern = NULL, *return_Value;
    int result = 1;
    int value, len;

    return_Value = NULL;

    if (!PyInt_Check(slice) && !PyTuple_Check(slice)) {
      PyErr_BadArgument();
      return -1;
    }

    /* it can be an integer */
#if PYTHON_API_VERSION >= 1007
    if (PyInt_Check(slice) && (value = PyInt_AS_LONG(slice)) == exp_end) {
#else
    if (PyInt_Check(slice) && (value = PyInt_AsLong(slice)) == exp_end) {
#endif
      ecase->type = value;
      return 0;
    } else if (PyInt_Check(slice)) {
      PyErr_SetString(PyExc_TypeError, "expecting tuple, got integer");
      return -1;
    }

    /* okay, we known it's a tuple */
    len = PyTuple_Size(slice);

    if ((obj = PySequence_GetItem(slice, 0)) == NULL)
      return -1;

    if (!PyInt_Check(obj)) {
      PyErr_SetString(PyExc_TypeError, "expecting number inside tuple");
      Py_DECREF(obj);
      return -1;
    }
#if PYTHON_API_VERSION >= 1007
    value = PyInt_AS_LONG(obj);
#else
    value = PyInt_AsLong(obj);
#endif
    Py_DECREF(obj);

    if (value == exp_end)
      result = 0;

    else {
      ecase->type = value;

      /* Determine the return value; this must be a hashable object.  */
      if ((return_Value = PySequence_GetItem(slice, len-1)) == NULL) {
        result = -1;
      } else if ((ecase->value = PyObject_Hash(return_Value)) == -1) {
        PyErr_SetString(PyExc_TypeError, "unhashable type");
        result = -1;
      } else {

        switch (value) {
          case exp_glob:
          case exp_exact:
            if (len != 3) {
              PyErr_SetString(PyExc_TypeError,
                "three values required in tuple");
              result = -1;
            } else {
              obj = PySequence_GetItem(slice, 1);
              if (PyString_Check(obj))
#if PYTHON_API_VERSION >= 1007
                ecase->pattern = PyString_AS_STRING(obj);
#else
                ecase->pattern = PyString_AsString(obj);
#endif
              else {
                PyErr_SetString(PyExc_TypeError, "must supply a string");
                result = -1;
              }
              Py_DECREF(obj);
            }
            break;

#ifdef HAVE_TCL
          case exp_regexp:
            if (len != 3) {
              PyErr_SetString(PyExc_TypeError,
                "three values required in tuple");
              result = -1;
            } else {
              obj = PySequence_GetItem(slice, 1);
              if (!PyString_Check(obj)) {
                PyErr_SetString(PyExc_TypeError, "must supply a string");
                result = -1;
              } else {
#if PYTHON_API_VERSION >= 1007
                ecase->pattern = PyString_AS_STRING(obj);
#else
                ecase->pattern = PyString_AsString(obj);
#endif
              }
              Py_DECREF(obj);
            }
            break;

          case exp_compiled:
            obj = PySequence_GetItem(slice, len-2);
            if (!ExpPyTclRE_Check(obj)) {
              PyErr_SetString(PyExc_TypeError,
                "must supply a regular expression");
              result = -1;
              pattern = NULL;
            } else {
              ecase->re = (TCL_REGEXP *)ExpPyTclRE_regexp(obj);
              pattern = obj;                  /* value to pass to callback */
              Py_INCREF(pattern);
              if (len == 4) {
                Py_DECREF(obj);
                obj = PyTuple_GetItem(slice, 1);
                if (PyString_Check(obj))
#if PYTHON_API_VERSION >= 1007
                  ecase->pattern = PyString_AS_STRING(obj);
#else
                  ecase->pattern = PyString_AsString(obj);
#endif
                else {
                  PyErr_SetString(PyExc_TypeError, "must supply a string");
                  result = -1;
                }
              } else if (len == 3)
                ecase->pattern = (char *)ExpPyTclRE_pattern(obj);
            }
            Py_DECREF(obj);
            break;

#else /* not HAVE_TCL */
#ifdef HAVE_PCRE
          case exp_regexp:
            if (len != 3) {
              PyErr_SetString(PyExc_TypeError,
                "three values required in tuple");
              result = -1;
            } else {
              /* change to a compiled string */
              ecase->type = exp_compiled;
              obj = PyTuple_GetItem(slice, 1);
              if (PyString_Check(obj)) {
#if PYTHON_API_VERSION >= 1007
                ecase->pattern = PyString_AS_STRING(obj);
#else
                ecase->pattern = PyString_AsString(obj);
#endif
                ecase->re = (regexp *)ExpPyRE_Create(obj);
              } else {
                PyErr_SetString(PyExc_TypeError, "must supply a string");
                result = -1;
              }
              Py_DECREF(obj);
            }
            break;

          case exp_compiled:
            obj = PySequence_GetItem(slice, len-2);
            if (!PyObject_HasAttrString(obj, "match")) {
              PyErr_SetString(PyExc_TypeError,
                "must supply a regular expression");
              result = -1;

            } else if ((ecase->re = (regexp *)ExpPyRE_New(obj)) == NULL)
              result = -1;

            else {
              pattern = (PyObject *)ExpPyRE_re((struct regexp *)ecase->re);
              Py_INCREF(pattern);
              if (len == 4) {
                Py_DECREF(obj);
                obj = PySequence_GetItem(slice, 1);
                if (PyString_Check(obj))
#if PYTHON_API_VERSION >= 1007
                  ecase->pattern = PyString_AS_STRING(obj);
#else
                  ecase->pattern = PyString_AsString(obj);
#endif
                else {
                  PyErr_SetString(PyExc_TypeError, "must supply a string");
                  result = -1;
                }
              } else if (len == 3) {
                Py_DECREF(obj);
                if ((obj = PyObject_GetAttrString(obj, "pattern")) == NULL)
                  result = -1;
                else if (PyString_Check(obj))
#if PYTHON_API_VERSION >= 1007
                  ecase->pattern = PyString_AS_STRING(obj);
#else
                  ecase->pattern = PyString_AsString(obj);
#endif
                else {
                  PyErr_SetString(PyExc_TypeError,
                    "must supply a regular expression");
                  result = -1;
                }

              } else {
                PyErr_SetString(PyExc_TypeError, "must supply a string");
                result = -1;
              }
            }
            Py_DECREF(obj);
            break;

#else /* not HAVE_PCRE */
          case exp_regexp:
          case exp_compiled:
            /* This is needed just to import the TclDummy.o module */
            (void)TclRegComp("");
            PyErr_SetString(Expect_Error, "regular expressions unsupported");
            result = -1;
            break;

#endif /* HAVE_PCRE */
#endif /* HAVE_TCL */
          case exp_null:
            ecase->pattern = "";  /* or should this be NULL? */
            break;

          default:
            PyErr_SetString(PyExc_ValueError,
    "Invalid type, must be END, GLOB, EXACT, REGEX, COMPILED, or NULL");
            result = -1;
        }
      }
    }

    if (result > 0) {
      cb_key = cb_item = NULL;
      if (pattern == NULL) {
        pattern = Py_None;
        Py_INCREF(pattern);
      }

      if ((cb_key = PyInt_FromLong(ecase->value)) != NULL &&
          (cb_item = Py_BuildValue("(OO)", return_Value, pattern)) != NULL) {
        /* there doesn't seem to be a "PyMapping_SetItem" function */
        PyDict_SetItem(callbacks, cb_key, cb_item);
      } else
        result = -1;
    }

    return result;
  }

static PyObject *
process_callback(callbacks, result, spawn_id)
  PyObject *callbacks;
  int result;
  ExpectObject *spawn_id;
  { PyObject *cb_key, *cb_item;
    PyObject *resultObj = NULL;
    PyObject *obj, *pattern;
    PyObject *match_obj, *cb_args;

    PyErr_Clear();

    if ((cb_key = PyInt_FromLong(result)) != NULL &&
        (cb_item = PyDict_GetItem(callbacks, cb_key)) != NULL) {
      PyArg_ParseTuple(cb_item, "OO", &obj, &pattern);
      if (PyCallable_Check(obj)) {
        if (pattern != Py_None) {
#ifdef HAVE_TCL
            match_obj = pattern;
#else
#ifdef HAVE_PCRE
            match_obj = PyObject_GetAttrString(pattern, "lastmatch");
            if (match_obj == NULL) {
              match_obj = Py_None;
              Py_INCREF(match_obj);
            }
#endif
#endif
            cb_args = Py_BuildValue("(OO)", ExpObj_match(spawn_id), match_obj);
        } else
            cb_args = Py_BuildValue("(O)", ExpObj_match(spawn_id));
        resultObj = PyObject_CallObject(obj, cb_args);

        Py_DECREF(cb_args);
      } else
        resultObj = obj;
    }

    Py_XDECREF(cb_key);
    return resultObj;
  }

/* create a new spawn object */
static PyObject *
new_ExpectObject(fd, argv)
  int fd;
  char *argv[];
  { ExpectObject *newobj;
    int buflen, l;
    char **arg, buffer[256];
    newobj = PyObject_NEW(ExpectObject, &ExpObj_Type);

    if (newobj) {
      ExpObj_id(newobj) = fd;
      ExpObj_pid(newobj) = exp_pid;
      ExpObj_closed(newobj) = 0;
      Py_INCREF(Py_None); ExpObj_buffer(newobj) = Py_None;
      Py_INCREF(Py_None); ExpObj_match(newobj) = Py_None;
      /* construct the command name */
      buflen = 0;
      (void)strcpy(buffer, "");
      for (arg = argv; *arg; arg++) {
        l = strlen(*arg);
        if (buflen + l + 1 < 256) {   /* this is to prevent memory overrun */
          (void)strcat(buffer, *arg);
          (void)strcat(buffer, " ");
          buflen += l + 1;
        /* we can not add more so add ellipsis */
        } else if (buflen + 4 < 256) {
          (void)strcat(buffer, "... ");
          buflen += 4;
        } else {
          (void)strcpy(&buffer[buflen-4], "... ");
        }
      }
      buffer[buflen-1] = '\0';  /* remove that last space */
      ExpObj_command(newobj) = PyString_FromString(buffer);
      ExpObj_pty(newobj) = PyString_FromString(exp_pty_slave_name);
      ExpObj_sttyobj(newobj) = Stty_new(fd);
    }
    return (PyObject *)newobj;
  }

/* create an unspawned object */
static PyObject *
unspawned_ExpectObject(fd, name)
  int fd;
  char *name;
  { ExpectObject *newobj;
    newobj = PyObject_NEW(ExpectObject, &ExpObj_Type);

    if (newobj) {
      exp_spawnfd(fd);
      ExpObj_id(newobj) = fd;
      ExpObj_pid(newobj) = 0;
      ExpObj_closed(newobj) = 0;
      Py_INCREF(Py_None); ExpObj_buffer(newobj) = Py_None;
      Py_INCREF(Py_None); ExpObj_match(newobj) = Py_None;
      ExpObj_command(newobj) = PyString_FromString("");
      ExpObj_pty(newobj) = PyString_FromString(name);
      ExpObj_sttyobj(newobj) = Stty_new(fd);
    }
    return (PyObject *)newobj;
  }

/* This might do something in the future.  The problem is that the Expect
   C library uses globals for a lot of situations.  The trick is finding
   out which we need.
 */
static int
Context_switch(self)
  ExpectObject *self;
  { int rc = 1;

    if (ExpObj_pid(self) != 0) {
      exp_pid = ExpObj_pid(self);
      if (!(rc = pyexp_alive(self->pid)))
        PyErr_SetString(Expect_Error, "Process exited prematurely");
    }
    return rc;
  }

/* close the object, this may mean that you only set the closed attribute
   to 1 - on unspawned objects, ownership of the file is assumed to be
   someone else, and no process is associated with it.
 */
static PyObject *
Expect_close(self, args)
  ExpectObject *self;
  PyObject *args;
  { long status;

    Context_switch(self); /* is this needed? */
    if (!PyArg_ParseTuple(args, ""))
      return NULL;
    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream already closed");
      return NULL;
    }
    if (ExpObj_pid(self) != 0)
      status = pyexp_close(ExpObj_id(self), ExpObj_pid(self));
    else
      status = 0;
    ExpObj_closed(self) = 1;
    if (status < 0) {
      PyErr_SetFromErrno(PyExc_SystemError);
      return NULL;
    }
    return PyInt_FromLong(status);
  }

/* set up to continue the next the expect clause, the next time around
 */
static char Expect_continue__doc__[] =
  "Continue within an expect clause.  Has no effect outside of the\n\
expect method.";
static PyObject *
Expect_continue(self, args)
  ExpectObject *self;
  PyObject *args;
  { PyObject *returnObj = NULL;

    if (PyArg_ParseTuple(args, "")) {
      ExpObj_continue(self) = 1;
      returnObj = Py_None;
      Py_INCREF(Py_None);
    }
    return returnObj;
  }

/* The expect method - the workhorse of the module.
   This uses Expect's exp_expectv() function directly.
   Return values are based on the following:
   + -1             - raise IOError with errno
   + EXP_EOF        - raise EOFError, set buffer in object
   + EXP_TIMEOUT    - raise TimeoutError
   + EXP_FULLBUFFER - raise FullBufferError, set buffer in object
   + anything else  - return either:
     - the integer passed in the clause for the matched pattern
     - the return value of the callback in the clause for the matched
       pattern
     both buffer and match are set
 */
static char Expect_expect__doc__[] =
  "Read from spawned object until expected contents is found.";
static PyObject *
Expect_expect(self, args)
  ExpectObject *self;
  PyObject *args;
  { int fd, i, num, size;
    struct exp_case *cases;
    int value, result;
    PyObject *obj, *resultObj;
    PyObject *slice; /* should be PyTupleObject, but I'll allow a raw EXP_END*/
    PyObject *cb_key, *cb_args;
    PyObject *callbacks;

    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream is closed");
      return NULL;
    }

    if (!Context_switch(self)) {
      PyErr_SetString(PyExc_EOFError, "End of stream reached");
      Py_DECREF(ExpObj_buffer(self));
      ExpObj_buffer(self) = Py_None;
      Py_INCREF(ExpObj_buffer(self));
      Py_DECREF(ExpObj_match(self));
      ExpObj_match(self) = Py_None;
      Py_INCREF(ExpObj_match(self));
      return NULL;
    }

    if ((callbacks = PyDict_New()) == NULL) {
      PyErr_NoMemory();
      return NULL;
    }

    num = PyTuple_Size(args);
    cases = new_exp_cases(num+1);  /* +1 in case END was not included */
    /* Clear out the exp_cases - this is so free_exp_cases() does get
       somewhere it shouldn't with non-NULL values.
       We also allow that exp_end is implied by the end of the
       argument list.
     */
    for (i=0; i < num+1; i++) {
      cases[i].type = exp_end;
      cases[i].value = 0;
      cases[i].re = NULL;
      cases[i].pattern = NULL;
    }

    for (i=0; i < num; i++) {
      if ((slice = PyTuple_GetItem(args, i)) == NULL) {
        free_exp_cases(cases, num);
        return NULL;
      }
      value = assign_exp_case(&cases[i], slice, callbacks);
      if (value == -1) {
        free_exp_cases(cases, num);
        return NULL;
      } else if (value == 0)
        break;
    }

    do {
      ExpObj_continue(self) = 0;  /* clear out the exp_continue flag */
      PyErr_Clear();
      fd = ExpObj_id(self);
      /* Thread safe? */
      Py_BEGIN_ALLOW_THREADS
      result = exp_expectv(fd, cases);
      Py_END_ALLOW_THREADS

      resultObj = NULL;
      if (PyErr_Occurred()) {  /* should this ever be set? */
        ;
      } else if (result == -1) {
        if (errno == EINTR) { /* system call interrupt */
          ExpObj_continue(self) = 1;
        } else
          PyErr_SetFromErrno(PyExc_SystemError);

      } else if (result == EXP_EOF) {
        PyErr_SetString(PyExc_EOFError, "End of stream reached");
        size = (int)(exp_buffer_end - exp_buffer);
        Py_DECREF(ExpObj_buffer(self));
        ExpObj_buffer(self) = PyString_FromStringAndSize(exp_buffer, size);
        Py_DECREF(ExpObj_match(self));
        ExpObj_match(self) = Py_None;
        Py_INCREF(ExpObj_match(self));

      } else if (result == EXP_TIMEOUT) {
        PyErr_SetString(Expect_TimeoutError, "Input timed out");
      } else if (result == EXP_FULLBUFFER) {
        PyErr_SetString(Expect_FullBufferError, "Input buffer is full");
        /* exp_buffer should be full */
        size = (int)(exp_buffer_end - exp_buffer);
        Py_DECREF(ExpObj_buffer(self));
        ExpObj_buffer(self) = PyString_FromStringAndSize(exp_buffer, size);
        Py_DECREF(ExpObj_match(self));
        ExpObj_match(self) = Py_None;
        Py_INCREF(ExpObj_match(self));
      } else {

        /* set the exp_buffer and exp_match */
        Py_DECREF(ExpObj_buffer(self));
        size = (int)(exp_buffer_end - exp_buffer);
        ExpObj_buffer(self) = PyString_FromStringAndSize(exp_buffer, size);
        Py_DECREF(ExpObj_match(self));
        size = (int)(exp_match_end - exp_match);
        ExpObj_match(self)  = PyString_FromStringAndSize(exp_match, size);

        if ((resultObj = process_callback(callbacks, result, self)) == NULL)
          /* exit the loop and propagate the exception */
          ExpObj_continue(self) = 0;

        /* when we loop around, this object will be lost but still
           referenced, dereference it now */
        else if (ExpObj_continue(self)) {
          Py_XDECREF(resultObj);
          resultObj = NULL;
        }

      }
    } while (ExpObj_continue(self));

    free_exp_cases(cases, num);
    Py_DECREF(callbacks);
    return resultObj;
  }

/* return the file descriptor, this will also keep it behaving like other
   file-like objects within Python
 */
static PyObject *
Expect_fileno(self, args)
  ExpectObject *self;
  PyObject *args;
  { int fd;
    PyObject *returnObj = NULL;

    if (!Context_switch(self))
      return NULL;

    if (!PyArg_ParseTuple(args, ""))
      return NULL;
    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream is closed");
      return NULL;
    }
    fd = ExpObj_id(self);
    returnObj = PyInt_FromLong(fd);
    return returnObj;
  }

/* return a Python file object representing the spawn object */
static char Expect_getfile__doc__[] =
  "Return a Python file object representing the spawned object.";
static PyObject *
Expect_getfile(self, args)
  ExpectObject *self;
  PyObject *args;
  { int fd;
    PyObject *returnObj = NULL;
    FILE *fp;

    if (!Context_switch(self))
      return NULL;

    if (!PyArg_ParseTuple(args, ""))
      return NULL;
    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream is closed");
      return NULL;
    }
    fd = ExpObj_id(self);
    Py_BEGIN_ALLOW_THREADS
    fp = fdopen(fd, "r+");
    Py_END_ALLOW_THREADS
    if (fp) {
      setbuf(fp, NULL);
      returnObj = PyFile_FromFile(fp, "<expect>", "r+", 0);
    }
    return returnObj;
  }

/* call the pyexp_interact function with the instance as the first argument */
static char Expect_interact__doc__[] =
  "Method shorthand for interact function.";
static PyObject *
Expect_interact(self, args)
  ExpectObject *self;
  PyObject *args;
  { PyObject *returnObj;
    PyObject *tuple, *input_spawn, *mappings;

    mappings = NULL;
    if (!Context_switch(self))
      return NULL;
    if (!PyArg_ParseTuple(args, "O|O", &input_spawn, &mappings))
      return NULL;
    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream is closed");
      return NULL;
    }

    if (mappings)
      tuple = Py_BuildValue("OOO", self, input_spawn, mappings);
    else
      tuple = Py_BuildValue("OO", self, input_spawn);
    if (tuple != NULL)
      returnObj = pyexp_interact(NULL, tuple);
      /* returnObj should only be NULL anyway */
    else
      returnObj = tuple;
    Py_XDECREF(tuple);
    return returnObj;  /* should be NULL */
  }

/* just write to the object's file descriptor */
static char Expect_send__doc__[] =
  "Send data to spawned process.";
static PyObject *
Expect_send(self, args)
  ExpectObject *self;
  PyObject *args;
  { int str_len, fd, rc;
    char *str_to_send;

    if (!Context_switch(self))
      return NULL;

    if (!PyArg_ParseTuple(args, "s#", &str_to_send, &str_len))
      return NULL;
    if (ExpObj_closed(self)) {
      PyErr_SetString(Expect_Error, "stream is closed");
      return NULL;
    }
    fd = ExpObj_id(self);
    Py_BEGIN_ALLOW_THREADS
    rc = write(fd, str_to_send, str_len);
    Py_END_ALLOW_THREADS
    if (rc == -1)
      return PyErr_SetFromErrno(PyExc_IOError);
    Py_INCREF(Py_None);
    return Py_None;
  }

/* the object's methods */
static struct PyMethodDef ExpectObj_methods[] = {
  { "close", (PyCFunction)Expect_close, METH_VARARGS },
  { "cont", (PyCFunction)Expect_continue, METH_VARARGS,
    Expect_continue__doc__ },
  { "expect", (PyCFunction)Expect_expect, METH_VARARGS,
    Expect_expect__doc__ },
  { "fileno", (PyCFunction)Expect_fileno, METH_VARARGS },
  { "getfile", (PyCFunction)Expect_getfile, METH_VARARGS,
    Expect_getfile__doc__ },
  { "interact", (PyCFunction)Expect_interact, METH_VARARGS,
    Expect_interact__doc__ },
  { "send", (PyCFunction)Expect_send, METH_VARARGS,
    Expect_send__doc__ },
  { NULL, NULL },
};

static char Exptype__doc__[] =
  "A spawned object";

/* the objects __del__ method */
static void
Expect_dealloc(self)
  ExpectObject *self;
  {
    if (!ExpObj_closed(self) && ExpObj_pid(self) != 0)
      pyexp_close(ExpObj_id(self), ExpObj_pid(self));
    PyMem_DEL(self);
  }

/* <open spawn 'command args' at xxxxxxxx> */
static PyObject *
Expect_repr(self)
  ExpectObject *self;
  { char buffer[256]; /* doesn't need to be this large */
    char ident[256];
    if (ExpObj_pid(self) == 0)
      sprintf(ident, "'%s'",
#if PYTHON_API_VERSION >= 1007
        PyString_AS_STRING(ExpObj_pty(self))
#else
        PyString_AsString(ExpObj_pty(self))
#endif
      );
    else if (ExpObj_closed(self))
      sprintf(ident, "'%s'",
#if PYTHON_API_VERSION >= 1007
        PyString_AS_STRING(ExpObj_command(self))
#else
        PyString_AsString(ExpObj_command(self))
#endif
      );
    else
      sprintf(ident, "'%s' (%d)",
#if PYTHON_API_VERSION >= 1007
        PyString_AS_STRING(ExpObj_command(self)),
#else
        PyString_AsString(ExpObj_command(self)),
#endif
        ExpObj_pid(self));
    sprintf(buffer, "<%s spawn %s at %p>",
            (ExpObj_closed(self)) ? "closed" : "open", ident, self);
    return PyString_FromString(buffer);
  }

/* all except the "id" field, which can be returned with the "fileno"
   method.
 */
static PyObject *
Expect_getattr(self, name)
  ExpectObject *self;
  char *name;
  {
    if (strcmp(name, "pid") == 0)
      if (ExpObj_pid(self) == 0) {   /* an unspawned object */
        Py_INCREF(Py_None);
        return Py_None;
      } else
        return PyInt_FromLong(ExpObj_pid(self));
    else if (strcmp(name, "command") == 0) {
      Py_INCREF(ExpObj_command(self));
      return ExpObj_command(self);
    } else if (strcmp(name, "pty") == 0) {
      Py_INCREF(ExpObj_pty(self));
      return ExpObj_pty(self);
    } else if (strcmp(name, "buffer") == 0) {
      Py_INCREF(ExpObj_buffer(self));
      return ExpObj_buffer(self);
    } else if (strcmp(name, "match") == 0) {
      Py_INCREF(ExpObj_match(self));
      return ExpObj_match(self);
    } else if (strcmp(name, "stty") == 0) {
      if (ExpObj_sttyobj(self) == NULL) {
        Py_INCREF(Py_None);
        return Py_None;
      } else {
        Py_INCREF(ExpObj_sttyobj(self));
        return ExpObj_sttyobj(self);
      }
    } else if (strcmp(name, "__members__") == 0) {
      return Py_BuildValue("[ssssss]",
        "buffer", "command", "match", "pid", "pty", "stty");
    } else
      return Py_FindMethod(ExpectObj_methods, (PyObject *)self, name);
  }

static PyTypeObject ExpObj_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,                              /*ob_size*/
  "spawn",                        /*tp_name*/
  sizeof(ExpectObject),           /*tp_basicsize*/
  0,                              /*tp_itemsize*/
  /* methods */
  (destructor)Expect_dealloc,     /*tp_dealloc*/
  (printfunc)0,                   /*tp_print*/
  (getattrfunc)Expect_getattr,    /*tp_getattr*/
  (setattrfunc)0,                 /*tp_setattr*/
  (cmpfunc)0,                     /*tp_compare*/
  (reprfunc)Expect_repr,          /*tp_repr*/
  0,                              /*tp_as_number*/
  0,                              /*tp_as_sequence*/
  0,                              /*tp_as_mapping*/
  (hashfunc)0,                    /*tp_hash*/
  (ternaryfunc)0,                 /*tp_call*/
  (reprfunc)0,                    /*tp_str*/

  /* Space for future expansion */
  0L,0L,0L,0L,
  Exptype__doc__                 /* Documentation string */
};


/* Expect Settings */
/* Create a Python object that sets and retrieves values from the
   Expect run-time environment.
   Currently these values include:
       ExpectPy   |    C library     |    Expect
   +--------------+------------------+--------------+
   | loguser      | exp_loguser      | loguser      |
   | internal     | exp_is_debugging | exp_internal |
   | timeout      | exp_timeout      | timeout      |
   | full_buffer  | exp_full_buffer  | full_buffer  |
   | disconnected | exp_disconnected | <none>       |
   | remove_nulls | exp_remove_nulls | remove_nulls |
   | tty_init     | exp_ttyinit      | -nottyinit   |
   | tty_copy     | exp_ttycopy      | -nottycopy   |
   | stty_init    | exp_stty_init    | stty_init    |
   +--------------+------------------+--------------+
   The factory_settings union allows for resetting to the values of the
   variables from the start of the program.  These values are set when
   the "settings" instance is created (module-init) not at real program
   start.
 */

/* until this gets out of hand, it's fast enough as a linear search
   then possibly as a binary tree
   Note: including exp_version support means that a lot of other
   variables get imported from Tcl libraries, which we probably
   don't want.
 */
static struct pyexp_variable_map {
  char *name;    /* name in Python */
  union {
    caddr_t addr;
    int *bool_value;
    int *int_value;
    char **str_value;
  } addr;        /* address of expect run-time variable */
  int type;      /* integer, string or boolean */
  int mod;       /* is it modifiable? */
  union {
    int bool_value;
    int int_value;
    char *str_value;
  } factory_settings;
} pyexp_variables[] = {

  { "loguser", (caddr_t)&exp_loguser, PYEXP_VAR_BOOL, 1 },
  { "internal", (caddr_t)&exp_is_debugging, PYEXP_VAR_BOOL, 1 },
  { "timeout", (caddr_t)&exp_timeout, PYEXP_VAR_INT, 1 },
  { "full_buffer", (caddr_t)&exp_full_buffer, PYEXP_VAR_BOOL, 1 },
  { "disconnected", (caddr_t)&exp_disconnected, PYEXP_VAR_BOOL, 0 },
  { "remove_nulls", (caddr_t)&exp_remove_nulls, PYEXP_VAR_BOOL, 1 },
#if 0
  /* this brings in all kinds of Tcl globals and functions that we
     don't necessarily want */
  { "version", (caddr_t)&exp_version, PYEXP_VAR_STR, 0 },
#endif
  { "tty_copy", (caddr_t)&exp_ttycopy, PYEXP_VAR_BOOL, 1 },
  { "tty_init", (caddr_t)&exp_ttyinit, PYEXP_VAR_BOOL, 1 },
  { "stty_init", (caddr_t)&exp_stty_init, PYEXP_VAR_STR, 1 },
  { NULL, NULL },

};

#define varmap_name(v)         ((v)->name)
#define varmap_bool(v)         ((v)->addr.bool_value)
#define varmap_int(v)          ((v)->addr.int_value)
#define varmap_str(v)          ((v)->addr.str_value)
#define varmap_type(v)         ((v)->type)
#define varmap_modifiable(v)   ((v)->mod)
#define varmap_fact_bool(v)    ((v)->factory_settings.bool_value)
#define varmap_fact_int(v)     ((v)->factory_settings.int_value)
#define varmap_fact_str(v)     ((v)->factory_settings.str_value)

/* return the number of items in the map; this could be more efficiently
   calculated with (sizeof(map)-1).
 */
static int
map_length(map)
  struct pyexp_variable_map *map;
  { int i;
    for (i=0; varmap_name(&map[i]); i++) ;
    return i;
  }

/* return the structure with the appropriate name (i.e. hide the
   organization and search algorithm).
 */
static struct pyexp_variable_map *
map_find(map, name)
  struct pyexp_variable_map *map;
  char *name;
  { struct pyexp_variable_map *item;

    for (item = map; varmap_name(item); item++)
      if (strcmp(varmap_name(item), name) == 0)
        return item;
    return NULL;
  }

/* create a new string and copy the contents of the given string into it,
   store that in the variable.  If to_free is non-zero, free that existing
   memory.  (There is some dispute about freeing the existing Expect
   string values.)
 */
static int
map_str_set(var, strval, to_free)
  struct pyexp_variable_map *var;
  char *strval;
  int to_free;
  { char *s;
    if (strval == NULL) {
      if (to_free && *varmap_str(var) != NULL)
        ExpPy_free(*varmap_str(var));
      *varmap_str(var) = strval;
      return 0;
    } else if ((s = ExpPy_calloc(char, strlen(strval)+1)) != NULL) {
      if (to_free && *varmap_str(var) != NULL)
        ExpPy_free(*varmap_str(var));
      strcpy(s, strval);
      *varmap_str(var) = s;
      return 0;
    } else
      return -1;
  }

/* store the existing Expect variable settings into the factory_settings.
 */
static void
map_factory_setting(var)
  struct pyexp_variable_map *var;
  {
    switch (varmap_type(var)) {
      case PYEXP_VAR_BOOL:
        varmap_fact_bool(var) = *varmap_bool(var);
        break;
      case PYEXP_VAR_INT:
        varmap_fact_int(var) = *varmap_int(var);
        break;
      case PYEXP_VAR_STR:
        /* what do we do with the Expect string - free it or leave it? */
        map_str_set(var, *varmap_str(var), 0);
        varmap_fact_str(var) = *varmap_str(var);
        break;
    }
  }

/* restore the factory_settings for an Expect variable */
static void
map_reset(var)
  struct pyexp_variable_map *var;
  {
    switch (varmap_type(var)) {
      case PYEXP_VAR_BOOL:
        *varmap_bool(var) = varmap_fact_bool(var);
        break;
      case PYEXP_VAR_INT:
        *varmap_int(var) = varmap_fact_int(var);
        break;
      case PYEXP_VAR_STR:
        *varmap_str(var) = varmap_fact_str(var);
        break;
    }
  }

/* it is not a complicated structure ;) */
typedef struct {
  PyObject_HEAD
  struct pyexp_variable_map *map;
} ExpSetObject;

staticforward PyTypeObject ExpSetObj_Type;
#define is_ExpSetObj(v)     ((v)->ob_type == &ExpSetObj_Type)
#define ExpSetObj_map(v)    (((ExpSetObject *)(v))->map)

/* inline documentation for the Python structure */
static char ExpSettype__doc__[] =
  "Access to expect variables\n\
  Name         Expect variable\n\
  ------------ ---------------\n\
  loguser      loguser\n\
  internal     exp_internal\n\
  timeout      timeout\n\
  full_buffer  full_nuffer\n\
  disconnected <N/A>\n\
  remove_nulls remove_nulls\n\
  tty_copy     spawn -nottycopy\n\
  tty_init     spawn -nottyinit\n\
  stty_init    stty_init\n\
Methods:\n\
  Name          Description\n\
  ------------  --------\n\
  reset([name]) reset all or one Expect variable to factory defaults\n\
  toDict()      convert Expect variables to a dict";

/* create a new Python "settings" object (all with the same data address)
 */
static PyObject *
new_ExpSetObject(map)
  struct pyexp_variable_map *map;
  { ExpSetObject *newobj;
    struct pyexp_variable_map *item;

    newobj = PyObject_NEW(ExpSetObject, &ExpSetObj_Type);
    if (newobj) {
      ExpSetObj_map(newobj) = map;
      for (item = map; varmap_name(item); item++)
        map_factory_setting(item);
    }
    return (PyObject *)newobj;
  }

/* method to reset one or all Expect variables registered with the
   settings object.  Variables are reset to values at module init time.
 */
static char ExpSet_reset__doc__[] =
 "reset([name]) - reset Expect variables to original values";
static PyObject *
ExpSet_reset(self, args)
  ExpSetObject *self;
  PyObject *args;
  { struct pyexp_variable_map *var;
    int i, l;
    char *name;

    if (PyArg_ParseTuple(args, "s", &name)) {
      if ((var = map_find(ExpSetObj_map(self), name)) == NULL) {
        PyErr_SetString(PyExc_AttributeError, name);
        return NULL;
      }
      map_reset(var);
    } else if (PyArg_ParseTuple(args, ""))
      for (var = ExpSetObj_map(self); varmap_name(var); var++)
        map_reset(var);
    else {
      PyErr_SetString(PyExc_TypeError, "reset requires as most 1 argument");
      return NULL;
    }

    Py_INCREF(Py_None);
    return Py_None;
  }

/* Convert the registered Expect variables to a Python dictionary.
   Changing the dictionary will not affect the Expect variables.
 */
static PyObject *
ExpSet_todict(self, args)
  ExpSetObject *self;
  PyObject *args;
  { struct pyexp_variable_map *var;
    PyObject *dict, *item;
    static PyObject *ExpSet_getattr(ExpSetObject *, char *);

    if ((dict = PyDict_New()) == NULL)
      return NULL;
    for (var = ExpSetObj_map(self); varmap_name(var); var++) {
      item = ExpSet_getattr(self, varmap_name(var));
      if (item == NULL) {
        Py_DECREF(dict);
        return NULL;
      }
      PyDict_SetItemString(dict, varmap_name(var), item);
    }
    return dict;
  }

static struct PyMethodDef ExpSetObj_methods[] = {
  { "reset", (PyCFunction)ExpSet_reset, METH_VARARGS, ExpSet_reset__doc__},
  { "toDict", (PyCFunction)ExpSet_todict, METH_VARARGS },
  { NULL, NULL },
};

/* Return the value of a register Expect variable as a Python object.
   Other possible attributes are:
     __members__  - list of Expect variable names
     __doc__      - inline documentation string
     __dict__     - (pre-1.5 only) a dictionary for the object (for use with
                    dir(settings)
 */
static PyObject *
ExpSet_getattr(self, varname)
  ExpSetObject *self;
  char *varname;
  { PyObject *returnval, *name;
    struct pyexp_variable_map *var;
    int i, size;

    if (strcmp(varname, "__members__") == 0) {
      size = map_length(ExpSetObj_map(self));
      /* +1 because of the "__doc__" member */
      if ((returnval = PyList_New(size + 1)) == NULL)
        return NULL;
      for (i=0; i < size; i++) {
        name = PyString_FromString(ExpSetObj_map(self)[i].name);
        PyList_SetItem(returnval, i, name);
      }
      PyList_SetItem(returnval, i, PyString_FromString("__doc__"));
    } else if (strcmp(varname, "__doc__") == 0) {
      returnval = PyString_FromString(ExpSettype__doc__);
#if PYTHON_API_VERSION < 1007
    /* needed so that 'dir(settings)' works; only in pre-1.5 */
    } else if (strcmp(varname, "__dict__") == 0) {
      returnval = ExpSet_todict(self, NULL);
#endif

    } else {
      var = map_find(ExpSetObj_map(self), varname);
      if (var && varmap_name(var)) {
        switch (varmap_type(var)) {
          case PYEXP_VAR_BOOL:
            if (*varmap_bool(var))
              returnval = PyInt_FromLong(1);
            else
              returnval = PyInt_FromLong(0);
            break;

          case PYEXP_VAR_INT:
            returnval = PyInt_FromLong(*varmap_int(var));
            break;

          case PYEXP_VAR_STR:
            if (*varmap_str(var) != NULL)
              returnval = PyString_FromString(*varmap_str(var));
            else {
              Py_INCREF(Py_None);
              returnval = Py_None;
            }
            break;

          default:
            PyErr_SetString(PyExc_TypeError, "Unsupported data type");
            return NULL;
        }
      } else {
        return Py_FindMethod(ExpSetObj_methods, (PyObject *)self, varname);
        /* NOTREACHED */
        PyErr_SetString(PyExc_AttributeError, varname);
        return NULL;
      }
    }
    PyErr_Clear();
    return returnval;
  }

/* Change the value of registered Expect variables with Python objects.
   Currently only boolean (zero or non-zero), integer or string values
   are allowed, or the results of __nonzero__, __int__, __str__ methods.
 */
static int
ExpSet_setattr(self, varname, value)
  ExpSetObject *self;
  char *varname;
  PyObject *value;
  { struct pyexp_variable_map *var;
    PyObject *temp;  /* used for "coersion" */
    char *str_val;
    int i, numb_val;

    if (value == NULL) { /* delattr */
      PyErr_SetString(PyExc_TypeError,
        "object doesn't support attribute deletion");
      return -1;
    }
    var = map_find(ExpSetObj_map(self), varname);
    if (var && varmap_name(var)) {
      if (!varmap_modifiable(var)) {
        PyErr_SetString(PyExc_SystemError, "read-only variable");
        return -1;
      } else

        switch (varmap_type(var)) {
          case PYEXP_VAR_BOOL:
            *varmap_bool(var) = PyObject_IsTrue(value);
            break;

          case PYEXP_VAR_INT:
            /* proper coersion should happen here */
            if (PyInt_Check(value)) {
#if PYTHON_API_VERSION >= 1007
              numb_val = PyInt_AS_LONG(value);
#else
              numb_val = PyInt_AsLong(value);
#endif
            } else if ((temp = PyNumber_Int(value)) != NULL) {
#if PYTHON_API_VERSION >= 1007
              numb_val = PyInt_AS_LONG(temp);
#else
              numb_val = PyInt_AsLong(temp);
#endif
              Py_DECREF(temp);
            } else {
              PyErr_SetString(PyExc_TypeError, "integer expected");
              return -1;
            }
            *varmap_int(var) = numb_val;
            break;

          case PYEXP_VAR_STR:
            if (PyString_Check(value)) {
#if PYTHON_API_VERSION >= 1007
              str_val = PyString_AS_STRING(value);
#else
              str_val = PyString_AsString(value);
#endif
            } else if (value == Py_None) {
              str_val = NULL;
            } else if ((temp = PyObject_Str(value)) != NULL) {
#if PYTHON_API_VERSION >= 1007
              str_val = PyString_AS_STRING(temp);
#else
              str_val = PyString_AsString(temp);
#endif
              Py_DECREF(temp);
            } else {
              PyErr_SetString(PyExc_TypeError, "string expected");
              return -1;
            }
            if (map_str_set(var, str_val, 1) < 0) {
              PyErr_NoMemory();
              return -1;
            }
            break;

          default:
            PyErr_SetString(PyExc_TypeError, "Unsupported data type");
            return -1;
        }
    } else if (strcmp(varname, "__doc__") == 0) {
      PyErr_SetString(PyExc_SystemError, "read-only variable");
      return -1;

    } else {
      PyErr_SetString(PyExc_AttributeError, varname);
      return -1;
    }
    PyErr_Clear();
    return 0;
  }

static PyTypeObject ExpSetObj_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,                              /*ob_size*/
  "settings",                     /*tp_name*/
  sizeof(ExpSetObject),           /*tp_basicsize*/
  0,                              /*tp_itemsize*/
  /* methods */
  (destructor)0,                  /*tp_dealloc*/
  (printfunc)0,                   /*tp_print*/
  (getattrfunc)ExpSet_getattr,    /*tp_getattr*/
  (setattrfunc)ExpSet_setattr,    /*tp_setattr*/
  (cmpfunc)0,                     /*tp_compare*/
  (reprfunc)0,                    /*tp_repr*/
  0,                              /*tp_as_number*/
  0,                              /*tp_as_sequence*/
  0,                              /*tp_as_mapping*/
  (hashfunc)0,                    /*tp_hash*/
  (ternaryfunc)0,                 /*tp_call*/
  (reprfunc)0,                    /*tp_str*/

  /* Space for future expansion */
  0L, 0L, 0L, 0L,
  ExpSettype__doc__               /* Documentation string */
};

/* the module's disconnect function... this just calls exp_disconnect() */
static char Exp_disconnect__doc__[] = "\
Disconnect from the controlling terminal.";
static PyObject *
pyexp_disconnect(self, args)
  PyObject *self, *args;
  {
    if (!PyArg_ParseTuple(args, ""))
      return NULL;
    Py_BEGIN_ALLOW_THREADS
    exp_disconnect();
    Py_END_ALLOW_THREADS
    Py_INCREF(Py_None);
    return Py_None;
  }

/* here for backward compatibility */
static char Exp_set__doc__[] = "\
Use settings object instead.  This is for backward compatibility only.";
static PyObject *
pyexp_set(self, args)
  PyObject *self, *args;
  { char *attr;
    PyObject *value, *returnval = NULL;

    /* name list */
    if (PyArg_ParseTuple(args, "")) {
      attr = "__members__";
      returnval = PyObject_GetAttrString(settings, attr);

    /* data storage */
    } else if (PyArg_ParseTuple(args, "sO", &attr, &value)) {
      returnval = PyObject_GetAttrString(settings, attr);
      if (returnval == NULL ||
          PyObject_SetAttrString(settings, attr, value) == -1)
        return NULL;

    /* data retrieval */
    } else if (PyArg_ParseTuple(args, "s", &attr)) {
      returnval = PyObject_GetAttrString(settings, attr);

    } else {
      PyErr_SetString(PyExc_TypeError,
        "Must supply at least a variable name");
      return NULL;
    }
    PyErr_Clear();
    return returnval;
  }

/* the module's spawn function, return a spawn object */
static char Exp_spawn__doc__[] = 
  "spawn(prog, arg0, [arg1, ...])\n\
Create an object tied to the spawned process.\n\
  prog      - program file to execute, if None use arg0\n\
  arg0      - process name\n\
  arg1, ... - arguments to program";
static PyObject *
pyexp_spawn(self, args)
  PyObject *self, *args;
  { int fd;
    PyObject *obj, *resultObj;
    char *file, *argv[256];  /* number based on most execv capabilities */
    int i, num;

    if ((num = PyTuple_Size(args)) < 2) {
      PyErr_SetString(PyExc_TypeError, "not enough arguments");
      return NULL;
    }

    if ((obj = PyTuple_GetItem(args, 0)) == NULL) {
      return NULL;
    }
    if (obj == Py_None)
      file = NULL;
    else if (PyString_Check(obj))
#if PYTHON_API_VERSION >= 1007
      file = PyString_AS_STRING(obj);
#else
      file = PyString_AsString(obj);
#endif
    else {
      PyErr_SetString(PyExc_TypeError, "expecting string");
      return NULL;
    }

    for (i=1; i < num; i++) {
      if ((obj = PyTuple_GetItem(args, i)) == NULL) {
        return NULL;
      }
      if (!PyString_Check(obj)) {
        PyErr_SetString(PyExc_TypeError, "expecting string");
        return NULL;
      }
#if PYTHON_API_VERSION >= 1007
      argv[i-1] = PyString_AS_STRING(obj);
#else
      argv[i-1] = PyString_AsString(obj);
#endif
    }
    argv[i-1] = NULL;

    if (file == NULL)
      file = argv[0];
    Py_BEGIN_ALLOW_THREADS
    fd = exp_spawnv(file, argv);
    Py_END_ALLOW_THREADS
    resultObj = new_ExpectObject(fd, argv);
    return resultObj;
  }

/* the module's spawnfile function */
static char Exp_spawn_fromfile__doc__[] = 
  "spawnfile(file)\n\
Create an object tied to a file-like object.";
static PyObject *
pyexp_spawn_fromfile(self, args)
  PyObject *self, *args;
  { int fd;
    PyObject *obj, *resultObj = NULL;
    PyObject *fileno, *fileno_meth, *value;
    char *name, *unknown = "<unknown>";

    if (PyArg_ParseTuple(args, "O", &obj)) {
      fileno_meth = NULL;
      if ((fileno_meth = PyObject_GetAttrString(obj, "fileno")) != NULL) {
        if ((fileno = PyEval_CallObject(fileno_meth, NULL)) != NULL) {
#if PYTHON_API_VERSION >= 1007
          fd = PyInt_AS_LONG(fileno);
#else
          fd = PyInt_AsLong(fileno);
#endif
          Py_DECREF(fileno);
          if ((value = PyObject_GetAttrString(obj, "name")) != NULL)
#if PYTHON_API_VERSION >= 1007
            name = PyString_AS_STRING(value);
#else
            name = PyString_AsString(value);
#endif

          else
            name = unknown;
          resultObj = unspawned_ExpectObject(fd, name);
          Py_XDECREF(value);
        }
      } else if (PyInt_Check(obj)) {
        resultObj = unspawned_ExpectObject(
#if PYTHON_API_VERSION >= 1007
                      PyInt_AS_LONG(value),
#else
                      PyInt_AsLong(value),
#endif
                      unknown);
      } else
        PyErr_SetString(PyExc_TypeError, "expecting file-like object");
      Py_XDECREF(fileno_meth);
    }

    return resultObj;
  }

/* module functions */
static PyMethodDef Expect_Methods[] = {
  { "disconnect", pyexp_disconnect, METH_VARARGS, Exp_disconnect__doc__},
  { "interact", pyexp_interact, METH_VARARGS, pyexp_interact__doc__ },
  { "set", pyexp_set, METH_VARARGS, Exp_set__doc__},
  { "spawn", pyexp_spawn, METH_VARARGS, Exp_spawn__doc__},
  { "spawnfile", pyexp_spawn_fromfile, METH_VARARGS,
    Exp_spawn_fromfile__doc__},
  { NULL, NULL }
};

/* create unspawn objects for standard input/output, for standard error,
   and for /dev/tty.
 */
static void
init_user_streams(module_dict)
  PyObject *module_dict;
  { PyObject *sysmod, *sysdict = NULL;
    PyObject *obj, *stream, *name;
    FILE *file;
    int fd;

    /* retrieve the streams from the sys module */
    if ((sysmod = PyImport_ImportModule("sys")) != NULL)
      sysdict = PyModule_GetDict(sysmod);
    Py_XDECREF(sysmod);

    if (sysdict != NULL) {
      stream = PyDict_GetItemString(sysdict, "stdin");
      if (PyFile_Check(stream)) {
        file = PyFile_AsFile(stream);
        fd = fileno(file);
        name = PyFile_Name(stream);
#if PYTHON_API_VERSION >= 1007
        obj = unspawned_ExpectObject(fd, PyString_AS_STRING(name));
#else
        obj = unspawned_ExpectObject(fd, PyString_AsString(name));
#endif
      } else {
        Py_INCREF(Py_None);
        obj = Py_None;
      }
      if (obj != NULL)
        PyDict_SetItemString(module_dict, "user_spawn_id", obj);

      stream = PyDict_GetItemString(sysdict, "stderr");
      if (PyFile_Check(stream)) {
        file = PyFile_AsFile(stream);
        fd = fileno(file);
        name = PyFile_Name(stream);
#if PYTHON_API_VERSION >= 1007
        obj = unspawned_ExpectObject(fd, PyString_AS_STRING(name));
#else
        obj = unspawned_ExpectObject(fd, PyString_AsString(name));
#endif
      } else {
        Py_INCREF(Py_None);
        obj = Py_None;
      }
      if (obj != NULL)
        PyDict_SetItemString(module_dict, "error_spawn_id", obj);

      Py_BEGIN_ALLOW_THREADS
      fd = open("/dev/tty", O_RDWR);
      Py_END_ALLOW_THREADS
      if (fd != -1)
        obj = unspawned_ExpectObject(fd, "tty");
      else {
        Py_INCREF(Py_None);
        obj = Py_None;
      }
      tty_spawn_id = obj;
      if (obj != NULL)
        PyDict_SetItemString(module_dict, "tty_spawn_id", obj);
    }

    Py_XDECREF(sysdict);
  }

/* Regular expression setup */
void
init_regexp_setup(dict)
  PyObject *dict;
  { PyObject *mod, *constant, *method;

    /* The value of "re_type" is the name of the module with regular
       expression support.
     */
#ifdef HAVE_TCL
    if ((constant = PyString_FromString("tclre")) != NULL) {
      PyDict_SetItemString(dict, "re_type", constant);
      Py_DECREF(constant);
    }
    initExpPyTcl();
    if ((mod = PyImport_ImportModule("tclre")) == NULL)
      Py_FatalError("Fatal Error: Cannot find tclre module");
    PyDict_SetItem(dict, constant, mod);
    if ((method = PyObject_GetAttrString(mod, "compile")) != NULL)
      PyDict_SetItemString(dict, "compile", method);
#else
#ifdef HAVE_PCRE
#if PYTHON_API_VERSION >= 1009
    if ((constant = PyString_FromString("pre")) != NULL) {
      PyDict_SetItemString(dict, "re_type", constant);
      Py_DECREF(constant);
    }
    initExpPyRE();
    if ((mod = PyImport_ImportModule("pre")) == NULL)
      Py_FatalError("Fatal Error: Cannot find pcre module");
    PyDict_SetItem(dict, constant, mod);
    if ((method = PyObject_GetAttrString(mod, "compile")) != NULL)
      PyDict_SetItemString(dict, "compile", method);
#else
    if ((constant = PyString_FromString("re")) != NULL) {
      PyDict_SetItemString(dict, "re_type", constant);
      Py_DECREF(constant);
    }
    initExpPyRE();
    if ((mod = PyImport_ImportModule("re")) == NULL)
      Py_FatalError("Fatal Error: Cannot find pcre module");
    PyDict_SetItem(dict, constant, mod);
    if ((method = PyObject_GetAttrString(mod, "compile")) != NULL)
      PyDict_SetItemString(dict, "compile", method);
#endif
#else
    Py_INCREF(Py_None);
    PyDict_SetItemString(dict, "re_type", Py_None);
#endif /* HAVE_PCRE */
#endif /* HAVE_TCL */
  }

void
initExpectPy()
  { PyObject *mod, *dict;
    PyObject *constant;

    /* By default, turn off logging (I find it really annoying) */
    exp_loguser = 0;

    mod = Py_InitModule("ExpectPy", Expect_Methods);
    dict = PyModule_GetDict(mod);

    /* version and copyright information */
    version = PyString_FromString(EXPPY_VERSION);
    PyDict_SetItemString(dict, "__version__", version);
    constant = PyString_FromString(copyright);
    PyDict_SetItemString(dict, "__copyright__", constant);
    settings = new_ExpSetObject(pyexp_variables);
    PyDict_SetItemString(dict, "settings", settings);

#if PYTHON_API_VERSION >= 1007
/* If class-based exceptions; don't know the specific macro tho */
    Expect_Error =
      PyErr_NewException("ExpectPy.error", NULL, NULL);
    Expect_TimeoutError =
      PyErr_NewException("ExpectPy.timeout", Expect_Error, NULL);
    Expect_FullBufferError =
      PyErr_NewException("ExpectPy.full_buffer", Expect_Error, NULL);
    PyDict_SetItemString(dict, "error", Expect_Error);
    PyDict_SetItemString(dict, "timeout", Expect_TimeoutError);
    PyDict_SetItemString(dict, "full_buffer", Expect_FullBufferError);
#else
    Expect_Error = Py_BuildValue("s", "expect.error");
    PyDict_SetItemString(dict, "error", Expect_Error);
    Expect_TimeoutError = Py_BuildValue("s", "expect.timeout");
    PyDict_SetItemString(dict, "timeout", Expect_TimeoutError);
    Expect_FullBufferError = Py_BuildValue("s", "expect.full_buffer");
    PyDict_SetItemString(dict, "full_buffer", Expect_FullBufferError);
#endif

    /* Set the constants */
    if ((constant = PyInt_FromLong(exp_end)) != NULL) {
      PyDict_SetItemString(dict, "END", constant);
      Py_DECREF(constant);
    }
    if ((constant = PyInt_FromLong(exp_glob)) != NULL) {
      PyDict_SetItemString(dict, "GLOB", constant);
      Py_DECREF(constant);
    }
    if ((constant = PyInt_FromLong(exp_exact)) != NULL) {
      PyDict_SetItemString(dict, "EXACT", constant);
      Py_DECREF(constant);
    }
    if ((constant = PyInt_FromLong(exp_regexp)) != NULL) {
      PyDict_SetItemString(dict, "REGEXP", constant);
      Py_DECREF(constant);
    }
    if ((constant = PyInt_FromLong(exp_compiled)) != NULL) {
      PyDict_SetItemString(dict, "COMPILED", constant);
      Py_DECREF(constant);
    }
    if ((constant = PyInt_FromLong(exp_null)) != NULL) {
      PyDict_SetItemString(dict, "NULL", constant);
      Py_DECREF(constant);
    }

    init_user_streams(dict);

    init_regexp_setup(dict);

  }

