#!/usr/local/bin/python

import sys
import posix
import strop

def exists(path):
  try:
    posix.stat(path)
  except OSError:
    return 0
  else:
    return 1
def is_dir(path):
  try:
    s = posix.stat(path)
  except OSError:
    return 0
  return (s[0] & 040000) == 040000
def is_executable(path):
  try:
    s = posix.stat(path)
  except OSError:
    return 0
  return (s[0] & 0111) != 0

def pathjoin(dir, base):
  return '%s/%s' % (dir, base)
def pathsplit(path):
  p = strop.rfind(path, '/')
  if p >= 0:
    return (path[:p], path[p+1:])
  return ('', path)
def splitext(path):
  p = strop.rfind(path, '.')
  if p >= 0:
    return (path[:p], path[p:])
  return (path, '')
def dirname(path):
  return pathsplit(path)[0]
def basename(path):
  return pathsplit(path)[1]

def ticks(*cmdargs):
  cmd = map(str, cmdargs)
  (stdout_rdr, stdout_wtr) = posix.pipe()

  pid = posix.fork()
  if pid == 0:
    posix.close(0); posix.open('/dev/null', posix.O_RDONLY)
    posix.close(1); posix.dup(stdout_wtr);
    posix.close(2); posix.dup(stdout_wtr);
    posix.close(stdout_wtr)
    posix.execv(cmd[0], cmd)
    posix.exit(255)

  else:
    posix.close(stdout_wtr)
    l = []
    d = posix.read(stdout_rdr, 1024)
    while d:
      l.append( d )
      d = posix.read(stdout_rdr, 1024)
    wpid = None
    while wpid != pid:
      (wpid, wstat) = posix.wait()
    v = strop.join(l)[:-1]
    if wstat:
      raise RuntimeError, v
    return v

def expect_expr(*args):
  global expect_prog
  cmd = 'puts [%s]' % strop.join(map(str, args))
  result = ticks(expect_prog, '-c', cmd)
  return result

def get_version(prog):
  version = expect_expr('exp_version')
  if strop.count(version, '.') > 1:
    p = strop.rfind(version, '.')
    return (version[:p], version)
  else:
    return (version, version)

# program start

try:
  forcea = posix.environ['FORCEA'] == '1'
except:
  forcea = 0

if len(sys.argv) > 1 and sys.argv[1] not in ('yes', 'no'):
  basedir = libdir = sys.argv[1]

  f = pathjoin(basedir, 'expect')
  if is_executable(f):
    expect_prog = f
  else:
    d = pathjoin(basedir, 'bin')
    f = pathjoin(d, 'expect')
    if is_executable(f):
      expect_prog = f
    else:
      raise SystemExit, "Cannot find expect interpreter"
    del d
    basevers, fullvers = get_version(expect_prog)
    libdir = pathjoin(basedir, pathjoin('lib', 'expect' + basevers))
  del f

else:
  if (posix.environ.has_key('EXPECT') and
      not is_executable(posix.environ['EXPECT'])):
    raise SystemExit, 'Cannot find expect interpreter'
  elif posix.environ.has_key('EXPECT'):
    expect_prog = posix.environ['EXPECT']
  else:
    path = strop.split(posix.environ['PATH'], ':')
    for dir in path:
      f = pathjoin(dir, 'expect')
      if is_executable(f):
        break
    else:
      raise SystemExit, 'Cannot find expect interpreter'
    expect_prog = f
    del path, dir, f

  basedir = expect_expr('set exp_library')
  libdir  = expect_expr('set exp_exec_library')

basevers, fullvers = get_version(expect_prog)

# find the include directory
incdir = pathjoin(basedir, 'include')
if is_dir(incdir) and exists(pathjoin(incdir, 'expect.h')):
  include_dir = incdir
elif exists(pathjoin(basedir, 'expect.h')):
  include_dir = basedir
else:
  raise SystemExit, 'Cannot find expect include files'
del incdir

minus_l_name = library_file = None

# find the libraries - get the latest version (based on filename order)
files = posix.listdir(libdir)
files.sort()
for file in files:
  if file[:9] == 'libexpect':
    base, ext = splitext(file)
    if ext not in ('.a', '.so', '.sl'):
      continue
    elif ext == '.a' and forcea:
      ext = '.a'
      library_file = pathjoin(libdir, file)
    elif ext in ('.so', '.sl') and not forcea:
      ext = '.so'
      library_file = pathjoin(libdir, file)
if library_file is not None:
  file, ext = splitext(basename(library_file))
  if file[-1:] == 'g' and file[9:-1] == basevers:
    minus_l_name=file[3:]  # remove 'lib'
  elif file[9:] == basevers:
    minus_l_name=file[3:]  # remove 'lib'
  else:
    pass
if minus_l_name is not None and not forcea:
  library_file = minus_l_name
library_dir = libdir

print "EXP_VERS='%s';" % basevers,
print "EXP_vers='%s';" % fullvers,
print "EXP_INCDIR='%s';" % include_dir,
print "EXP_LIBDIR='%s';" % library_dir,
if library_file is not None:
  print "EXP_LIB='%s';" % library_file,
print
